"""
A tool to drive Excel using the Capital IQ plugin to download Capital IQ data. Useful for downloading
large data sets from Capital IQ.
"""
from capiq_excel.main import (
    download_data,
    download_data_for_capiq_ids
)


from typing import Optional
from capiq_excel.tools.dates import freq_and_periods_to_begin_date_str, today_as_str


def financial_data_command(company_id: str, data_item: str, freq: str='Q', num_periods: int=80,
                           data_item_label: Optional[str]=None) -> str:
    """
    Lower-level utility to create an Excel function from inputs for getting financial data

    :param company_id: Capital IQ id, e.g. IQ21835
    :param data_item: Capital IQ item, e.g. IQ_TOTAL_REV. Look them up in Data -> Formula Builder in the Capital IQ
        Excel plugin tab
    :param freq: One character to represent frequency, Q or Y
    :param num_periods: Number of periods to go back in time pulling data
    :param data_item_label: Column name to use instead of data_item name
    :return: Excel command
    """
    _validate_financial_data_inputs(company_id, data_item, freq=freq, num_periods=num_periods, data_item_label=data_item_label)

    # Use variable name as label if none specified
    if data_item_label is None:
        data_item_label = data_item

    return f'=CIQRANGE("{company_id}", "{data_item}", IQ_F{freq} - {num_periods}, , , , , , "{data_item_label}")'


def market_data_command(company_id: str, data_item: str, freq: str='Q', num_periods: int=80,
                        data_item_label: Optional[str]=None) -> str:
    """
    Lower-level utility to reate an Excel function from inputs for getting market data

    For market data, Capital IQ is expecting begin and end date to be passed rather than relative date. This function
    converts the relative date to absolute dates then creates the Excel command.

    :param company_id: Capital IQ id, e.g. IQ21835
    :param data_item: Capital IQ item, e.g. IQ_FLOAT_PERCENT. Look them up in Data -> Formula Builder in the Capital IQ
        Excel plugin tab
    :param freq: One character to represent frequency, Q or Y
    :param num_periods: Number of periods to go back in time pulling data
    :param data_item_label: Column name to use instead of data_item name
    :return: Excel command
    """
    begin_date = freq_and_periods_to_begin_date_str(freq, num_periods)
    end_date = today_as_str()

    # Use variable name as label if none specified
    if data_item_label is None:
        data_item_label = data_item

    return f'=CIQRANGE("{company_id}", "{data_item}", "{begin_date}", "{end_date}", , , , , "{data_item_label}")'


def holdings_command(company_id, data_item, date_str, data_item_label=None):
    # Use variable name as label if none specified
    if data_item_label is None:
        data_item_label = data_item

    return f'=CIQRANGE("{company_id}", "{data_item}", 1, 50000, "{date_str}", , , , "{data_item_label}")'

def id_command(search_str):
    return f'=CIQRANGEA("{search_str}","IQ_COMPANY_ID_QUICK_MATCH",1,1)'

def name_command(search_str):
    return f'=CIQRANGEA("{search_str}","IQ_COMPANY_NAME_QUICK_MATCH",1,1)'

def _validate_financial_data_inputs(*args, **kwargs):
    assert kwargs['freq'] in ('Q','Y')
    assert isinstance(args[0], str)
    assert isinstance(args[1], str)


import time
from ..exceptions import WorkbookClosedException, CapitalIQInactiveException

def _wait_for_capiq_result(excel):
    finished = False
    while not finished:
        time.sleep(1)
        finished = _capiq_is_done(excel)

    return True

def _capiq_is_done(excel):
    if _ciq_not_working(excel):
        raise CapitalIQInactiveException('CapitalIQ Plugin did not properly pull values.')

    return _data_in_a2(excel)

def _ciq_not_working(excel):
    a1 = _get_cell_value_by_index(excel, 1, 1)
    if a1 is None:
        return True  # should always have something in A1
    error_values = ('refresh', 'ciqinactive', 'error')
    return any([value in a1.lower() for value in error_values])


def _data_in_a2(excel):
    a2 = _get_cell_value_by_index(excel, 1, 2)
    return a2 is not None

def _get_cell_value_by_index(excel, x, y):
    try:
        result = excel.ActiveSheet.Cells(1, 2).Value
        return result
    except AttributeError:
        raise WorkbookClosedException('Workbook was not open when trying to populate values.')


import re
import pandas as pd
import numpy as np
from exceldriver.columns import excel_cols


CIQ_MARKET_ITEM_FORMULA_PATTERN = re.compile(r'=CIQ\("[\w]+", "[\w]+", "([\w\/]+)"\)')


def extract_capiq_df_from_sheet(ws, market_data_items):
    col_gen = excel_cols()

    all_series = []
    while True:
        col = next(col_gen)
        if col == 'A':
            continue  # A is date column, don't need to extract
        cell_range = f'{col}1'
        value = ws.Range(cell_range).Value
        if value is None:
            # Blank column label, data is finished
            break
        if value in market_data_items.values():
            # Market data item, need to extract data
            all_series.append(get_series_from_ciq_market_item_col(ws, col))
        else:
            # Financial data item, need to extract data
            all_series.append(get_series_from_ciq_financial_col(ws, col))

    df = pd.concat(all_series, axis=1)
    df.index.name = 'Date'
    return df


def get_series_from_ciq_market_item_col(ws, col: str):
    series = pd.Series()
    row = 2
    while True:
        cell_range = f'{col}{row}'
        value = ws.Range(cell_range).Value
        if value is None:
            break
        formula = ws.Range(cell_range).Formula
        date = get_date_from_ciq_market_item_formula(formula)
        series[date] = value
        row += 1
    series.name = ws.Range(f'{col}1').Value
    series.index = pd.to_datetime(series.index)
    return series


def get_series_from_ciq_financial_col(ws, col: str):
    series = pd.Series()
    row = 2
    while True:
        cell_range = f'{col}{row}'
        value = ws.Range(cell_range).Value
        if value is None:
            break
        # Column A has dates
        date = ws.Range(f'A{row}').Value
        if isinstance(date, str):
            # Handle CapIQ missing representation
            if date == 'NA':
                date = np.nan
            # Allow other strs to pass through
        else:
            # TODO: add check for pycom datetime object type
            # Extract date str from pycom datetime object
            date = str(date.date())
        series[date] = value
        row += 1
    series.name = ws.Range(f'{col}1').Value
    series.index = pd.to_datetime(series.index)
    return series


def get_date_from_ciq_market_item_formula(formula: str) -> str:
    match = CIQ_MARKET_ITEM_FORMULA_PATTERN.match(formula)
    if not match:
        return None
    return match.group(1)


from typing import Dict
import pandas as pd
import warnings
from concurrent.futures import ThreadPoolExecutor, TimeoutError

from capiq_excel.downloader.timeout import TimeoutWrapper
from capiq_excel.fileops import get_path_of_failed_folder_add_if_necessary, move_file_to_failed_folder, get_path_of_additional_failed_folder_add_if_necessary
from capiq_excel.workbook.populate.main import populate_capiq_for_file
from exceldriver.tools import _start_excel_with_addins_and_attach, _get_excel_running_workbook, _restart_excel_with_addins_and_attach, NoExcelWorkbookException
from processfiles.files import FileProcessTracker



def populate_all_files_in_folder(folder, financial_data_items_dict: Dict[str, str],
                                 market_data_items_dict: Dict[str, str], restart=True, timeout=240,
                                 run_failed=False,):
    """

    """
    _validate_populate_inputs(folder, restart, run_failed)

    excel = _start_excel_with_addins_and_attach()
    failed_folder = get_path_of_failed_folder_add_if_necessary(folder)

    if run_failed:
        # Set main folder as 'failed', then set failed folder as another failed folder inside the original
        folder = failed_folder
        failed_folder = get_path_of_additional_failed_folder_add_if_necessary(folder)

    file_tracker = FileProcessTracker(folder=folder, restart=restart, file_types=('xlsx',))

    with ThreadPoolExecutor(max_workers=1) as e:
        for i, file in enumerate(file_tracker.file_generator()):

            excel, successful = _try_to_get_result_if_fail_restart_excel(
                e,
                i,
                file,
                excel,
                financial_data_items_dict=financial_data_items_dict,
                market_data_items_dict=market_data_items_dict
            )

            if not successful:
                move_file_to_failed_folder(file, failed_folder)


def _get_company_id_list(id_filepath, id_col='IQID'):
    df = pd.read_csv(id_filepath, usecols=[id_col])
    unique = df[id_col].unique().tolist()

    # Drop nans
    return [i for i in unique if not pd.isnull(i)]

def _validate_populate_inputs(folder, restart, run_failed):
    assert not (restart and run_failed)

    if run_failed:
        warnings.warn(f'run_failed flag passed. Folder {folder}, will not be run, instead the failed folder will')

### Functions below to assist with multiprocessing/timeout handling

def _try_to_get_result_if_fail_restart_excel(threadpool, i, file, excel, tries_remaining=3, **kwargs):

    if tries_remaining <= 0:
        print(fr'ERROR: Timed out processing {file}. Skipping and moving to "..\failed".')
        return excel, False

    # Submit result on threadpool asynchronously
    async_result = threadpool.submit(populate_capiq_for_file, file, excel, index=i + 1, **kwargs)

    # Try to get the result, waiting up to two minutes.
    try:
        excel, successful = async_result.result(timeout=300)
    except TimeoutError:
        excel = _restart_excel_with_addins_and_attach()
        return _try_to_get_result_if_fail_restart_excel(threadpool, i, file, excel, tries_remaining=tries_remaining - 1)


    return excel, successful


def _restart_excel_and_return_false():
    excel = _restart_excel_with_addins_and_attach(max_retries=10)

    return False

def _return_false():
    return False

## END TEMP


from typing import Sequence, List
import pandas as pd

from exceldriver.tools import _start_excel_with_addins_and_attach
from processfiles.files import FileProcessTracker
from capiq_excel.workbook.populate.main import populate_capiq_ids_for_file
from capiq_excel.workbook.create import create_all_xlsx_with_id_commands


def download_capiq_ids(ids: Sequence[str], outpath: str = 'capiq ids.csv', folder: str = 'in_process_ids') -> List[str]:
    """
    Downloads Capital IQ identifiers when passed other identifiers such as CUSIP,
    ISIN, ticker, name, etc.

    Stores in a CSV with matched names included and also returns capiq ids as a list

    :param ids: identifiers such as CUSIP, ISIN, ticker, name. Can be a mixture.
    :param folder: folder which will hold in process files
    :param outpath: filepath to output csv, including file extension
    :return: capiq ids
    """
    print('Creating XLSX files with commands to get ids')
    create_all_xlsx_with_id_commands(ids, folder, num_files=100)

    print('Populating XLSX files for ids')
    populate_all_ids_in_folder(folder)

    print('Combining all ids into a single CSV file')
    combine_all_capiq_ids_xlsx(folder, outpath)

    return _get_ids_from_csv_path(outpath)


def populate_all_ids_in_folder(folder, restart=True):
    excel = _start_excel_with_addins_and_attach()
    file_tracker = FileProcessTracker(folder=folder, restart=restart, file_types=('xlsx',))

    for file in file_tracker.file_generator():
        populate_capiq_ids_for_file(file, excel)


def combine_all_capiq_ids_xlsx(infolder, outpath, restart=True):

    df = pd.DataFrame()

    file_tracker = FileProcessTracker(folder=infolder, restart=restart, file_types=('xlsx',))

    for file in file_tracker.file_generator():
        df = _append_capiq_xlsx_to_df(df, file)

    _remove_useless_cols(df)
    df.to_csv(outpath, index=False)

    return df


def _append_capiq_xlsx_to_df(df, filepath):
    temp_df = pd.read_excel(filepath)
    df = df.append(temp_df)

    return df


def _remove_useless_cols(df):
    """
    NOTE: inplace
    """
    blank_cols = [col for col in df.columns if 'blank' in col.lower()]
    useless_cols = ['ID'] + blank_cols
    df.drop(useless_cols, axis=1, inplace=True)


def _get_ids_from_csv_path(csv_path: str) -> List[str]:
    df = pd.read_csv(csv_path)
    return df['IQID'].tolist()



import sys
import time
import traceback
from typing import Dict

import pythoncom
from win32com.client import constants
from pywintypes import com_error

from exceldriver.tools import _restart_excel_with_addins_and_attach
from exceldriver.columns import get_n_cols_after_col
from capiq_excel.workbook.wait import _wait_for_capiq_result
from capiq_excel.exceptions import WorkbookClosedException, CapitalIQInactiveException
from capiq_excel.workbook.populate.extract import extract_capiq_df_from_sheet
from capiq_excel.workbook.populate.replace import write_df_to_ws_values



def populate_capiq_for_file(filepath, excel, financial_data_items_dict: Dict[str, str],
                            market_data_items_dict: Dict[str, str], retries_remaining=3, close_workbook=False, index=0):
    """

    Private function has main functionality. This is a wrapper to add retries afer com errors
    """

    # Necessary to be called in each new thread or process which uses COM (communicate with Microsoft products)
    pythoncom.CoInitialize()

    # Even if things are going normally, restart every 500 worksheets as there is a memory leak
    if index % 500 == 0 and retries_remaining == 3:
        excel = _restart_excel_with_addins_and_attach()

    # Stop retries
    if retries_remaining <= 0:
        print(fr'ERROR: Could not process {filepath}. Skipping and moving to "..\failed".')
        return excel, False

    try:
        # If we are retrying, need to close the workbook before trying to populate
        if close_workbook:
            excel.CutCopyMode = False
            time.sleep(1)
            if excel.ActiveWorkbook:
                excel.ActiveWorkbook.Close(SaveChanges=False)
            time.sleep(5)

        _populate_capiq_for_file(filepath, excel, financial_data_items_dict, market_data_items_dict)
        return excel, True
    except (com_error, WorkbookClosedException, CapitalIQInactiveException) as e:
        print(f'Error {e} populating {filepath}. Will wait 30 seconds, restart Excel, and try again.')
        traceback.print_tb(sys.exc_info()[2])
        time.sleep(30)
        excel = _restart_excel_with_addins_and_attach()
        return populate_capiq_for_file(
            filepath,
            excel,
            financial_data_items_dict,
            market_data_items_dict,
            retries_remaining=retries_remaining - 1,
            close_workbook=True
        )


def _populate_capiq_for_file(filepath, excel, financial_data_items_dict: Dict[str, str],
                            market_data_items_dict: Dict[str, str]):
    wb = excel.Workbooks.Open(filepath)
    successful = _wait_for_capiq_result(excel)
    _set_date_format(excel, wb, cell_range='A:A')  # column A is automatically included date
    _extract_unaligned_data_align_and_replace(wb, market_data_items_dict)
    excel.ActiveWorkbook.Close(SaveChanges=True)
    return successful


def populate_capiq_ids_for_file(filepath, excel):
    wb = excel.Workbooks.Open(filepath)
    successful = _wait_for_capiq_result(excel)
    _copy_paste_values(excel, wb, cell_range='A:Z')
    excel.ActiveWorkbook.Close(SaveChanges=True)
    return successful


def _copy_paste_values(excel, wb, cell_range='A1:ZZ20000'):
    ws = wb.Sheets('Sheet')
    ws.Range(cell_range).Copy()
    ws.Range(cell_range.split(':')[0]).PasteSpecial(Paste=constants.xlPasteValues, Operation=constants.xlNone)
    excel.CutCopyMode = False


def _set_date_format(excel, wb, cell_range='B:B'):
    ws = wb.Sheets('Sheet')
    ws.Range(cell_range).NumberFormat = 'mm/dd/yyyy'


def _extract_unaligned_data_align_and_replace(wb, market_data_items_dict: Dict[str, str]):
    """
    Market data items become unaligned with the date axis. Extract the date from each individual formula,
    then combine everything and replace the contents of the worksheet
    """
    ws = wb.Sheets('Sheet')
    df = extract_capiq_df_from_sheet(ws, market_data_items_dict)
    write_df_to_ws_values(df, ws)



from typing import List, Dict, Union, Sequence, Optional
import os
from capiq_excel.workbook.create import create_all_xlsx_with_commands
from capiq_excel.downloader.tools import populate_all_files_in_folder
from capiq_excel.ids import download_capiq_ids, _get_ids_from_csv_path
from capiq_excel.combine import combine_all_capiq_xlsx


def download_data(company_ids: List[str], financial_data_items: Optional[Union[Dict[str, str], Sequence[str]]] = None,
                  market_data_items: Optional[Union[Dict[str, str], Sequence[str]]] = None,
                  ids_folder: str = 'in_process_ids', data_folder: str = 'in_process_data',
                  data_outpath: str = 'capiq data.csv', ids_outpath: str = 'capiq ids.csv',
                  restart: bool = True, timeout: int = 240, run_failed: bool = False,
                  **financial_command_kwargs):
    """
    Downloads data from Capital IQ given arbitrary ids such as name, ticker, CUSIP, ISIN, etc.

    :param company_ids:
    :param financial_data_items: Financial data items to pull for companies. Only pass variables which show up on the
        Financial Data tab in the formula builder. keys are variable names in capital iq, values are what the
        names should be in the output data. e.g. {'IQ_TOTAL_REVENUE': 'Sales'}. If wanting to use the original names
        from Capital IQ, simply a list may be passed, e.g. ['IQ_TOTAL_REVENUE']
    :param market_data_items: Market data items to pull for companies. Only pass variables which show up on the Market
        Data tab in the formula builder. Format is the same as financial_data_items
    :param ids_folder: Location for in-process files for getting ids
    :param data_folder: Location for in-process files for getting data
    :param data_outpath: Path to csv where final data should be outputted
    :param ids_outpath: Path to csv where Capital IQ ids should be outputted
    :param restart: If not the first run, start over from the beginning. Set to False to start from where left off
    :param timeout: Time to wait for file to be populated before considering it failed
    :param run_failed: Should only be set to True on a second or later run. Will target the failed files instead
        of the main files if True is passed.
    :param financial_command_kwargs: kwargs for :py:func:`.financial_data_command`
    :return:
    """

    if restart or not os.path.exists(ids_outpath):
        capiq_ids = download_capiq_ids(
            company_ids,
            outpath=ids_outpath,
            folder=ids_folder
        )
    else:
        capiq_ids = _get_ids_from_csv_path(ids_outpath)

    download_data_for_capiq_ids(
        capiq_ids,
        financial_data_items=financial_data_items,
        market_data_items=market_data_items,
        folder=data_folder,
        outpath=data_outpath,
        restart=restart,
        timeout=timeout,
        run_failed=run_failed,
        **financial_command_kwargs
    )


def download_data_for_capiq_ids(capiq_company_ids: List[str],
                                financial_data_items: Optional[Union[Dict[str, str], Sequence[str]]] = None,
                                market_data_items: Optional[Union[Dict[str, str], Sequence[str]]] = None,
                                folder: str = 'in_process_data',
                                outpath: str = 'capiq data.csv',
                                restart: bool = True, timeout: int = 240, run_failed: bool = False,
                                **financial_command_kwargs):
    """
    Downloads data from Capital IQ given the capital IQ ids and chosen variables

    :param capiq_company_ids: Capital IQ ids for companies, e.g. ['IQ4564656', 'IQ45643215']
    :param financial_data_items: Financial data items to pull for companies. Only pass variables which show up on the
        Financial Data tab in the formula builder. keys are variable names in capital iq, values are what the
        names should be in the output data. e.g. {'IQ_TOTAL_REVENUE': 'Sales'}. If wanting to use the original names
        from Capital IQ, simply a list may be passed, e.g. ['IQ_TOTAL_REVENUE']
    :param market_data_items: Market data items to pull for companies. Only pass variables which show up on the Market
        Data tab in the formula builder. Format is the same as financial_data_items
    :param folder: Location for in-process files
    :param outpath: Path to csv where final data should be outputted
    :param restart: If not the first run, start over from the beginning. Set to False to start from where left off
    :param timeout: Time to wait for file to be populated before considering it failed
    :param run_failed: Should only be set to True on a second or later run. Will target the failed files instead
        of the main files if True is passed.
    :param financial_command_kwargs: kwargs for :py:func:`.financial_data_command`
    :return:
    """
    financial_data_items, market_data_items = _get_data_items_dicts(financial_data_items, market_data_items)

    if restart or not os.path.exists(folder):
        print('Creating XLSX files with commands')
        create_all_xlsx_with_commands(
            folder,
            company_id_list=capiq_company_ids,
            financial_data_items_dict=financial_data_items,
            market_data_items_dict=market_data_items,
            **financial_command_kwargs
        )

    print('Populating XLSX files with Capital IQ data')
    populate_all_files_in_folder(
        folder,
        financial_data_items,
        market_data_items,
        restart=restart,
        timeout=timeout,
        run_failed=run_failed,
    )

    print('Combining individual XLSX files into a single CSV file')
    combine_all_capiq_xlsx(
        folder,
        outpath,
        restart=restart
    )


    # TODO: remove after fixing bug where download_data_for_capiq_ids combine isn't working on first go
    if not os.path.exists(outpath):
        combine_all_capiq_xlsx(
            folder,
            outpath
        )

def _get_data_items_dicts(financial_data_items: Union[Dict[str, str], Sequence[str]] = None,
                          market_data_items: Union[Dict[str, str], Sequence[str]] = None
                          ) -> List[Dict[str, str]]:
    if financial_data_items is None and market_data_items is None:
        raise ValueError('must pass either financial_data_items or market_data_items')
    out_dicts = []
    for data_items in (financial_data_items, market_data_items):
        if data_items is None:
            out_dicts.append({})
        elif not isinstance(data_items, dict):
            # Convert list to dict
            out_dicts.append({var_name: var_name for var_name in data_items})
        else:
            # Got dict already, just add to output unchanged
            out_dicts.append(data_items)
    return out_dicts


# TODO: download_ids, download_holdings, or combine those into the above function


from typing import Sequence, Dict, Iterable, Callable
import os
import string
import itertools
import math
import pandas as pd
from openpyxl.utils.dataframe import dataframe_to_rows

from exceldriver.workbook.create import get_workbook_and_worksheet
from exceldriver.columns import excel_cols
from .commands import financial_data_command, id_command, name_command, holdings_command, market_data_command


def create_all_xlsx_with_commands(folder: str, company_id_list: Sequence[str],
                                  financial_data_items_dict: Dict[str, str],
                                  market_data_items_dict: Dict[str, str], **financials_kwargs):
    [
        create_xlsx_with_commands(
            folder,
            company_id,
            financial_data_items_dict,
            market_data_items_dict,
            **financials_kwargs
        )
        for company_id in company_id_list
    ]


def create_xlsx_with_commands(folder: str, company_id: str, financial_data_items_dict: Dict[str, str],
                              market_data_items_dict: Dict[str, str], **financials_kwargs):
    wb, ws = get_workbook_and_worksheet()
    _fill_with_commands(ws, company_id, financial_data_items_dict, market_data_items_dict, **financials_kwargs)

    if not os.path.exists(folder):
        os.makedirs(folder)

    filepath = os.path.join(folder, f'{company_id}.xlsx')
    wb.save(filepath)

    return os.path.abspath(filepath)

def create_all_xlsx_with_holdings_commands(folder, company_id_list, date_str_list, data_items_dict):
    [
        create_xlsx_with_holdings_commands(folder, company_id, date_str, data_items_dict)
        for company_id, date_str in itertools.product(company_id_list, date_str_list)
    ]


def create_xlsx_with_holdings_commands(folder, company_id, date_str, data_items_dict):
    wb, ws = get_workbook_and_worksheet()
    _fill_with_holdings_commands(ws, company_id, date_str, data_items_dict)

    filepath = os.path.join(folder, f'{company_id} {_date_str_to_file_format(date_str)}.xlsx')
    wb.save(filepath)

    return os.path.abspath(filepath)

def create_all_xlsx_with_id_commands(ids: Sequence[str], folder, num_files=100):
    wb, ws = get_workbook_and_worksheet()

    if not os.path.exists(folder):
        os.makedirs(folder)

    df = pd.DataFrame()
    _fill_id_column(df, ids)
    _fill_capiq_id_column(df)
    _fill_capiq_name_column(df)

    rows_per_df = math.ceil(len(df)/num_files)

    count_per_wb = 0
    count_of_wb = 0
    for index, r in enumerate(dataframe_to_rows(df, index=False, header=True)):
        if index == 0:
            headers = r
        count_per_wb += 1
        ws.append(r)
        if count_per_wb >= rows_per_df:
            count_per_wb = 0
            count_of_wb += 1
            wb, ws = _save_wb_by_index_get_new_wb(count_of_wb, folder, wb)
            ws.append(headers)

##### Helper functions ####

def _save_wb_by_index_get_new_wb(index, folder, wb):
    filename = f'ids {index}.xlsx'
    filepath = os.path.join(folder, filename)
    wb.save(filepath)
    wb, ws = get_workbook_and_worksheet()
    return wb, ws

def _fill_id_column(df: pd.DataFrame, ids: Sequence[str]):
    """
    NOTE: inplace
    """
    df['ID'] = ids

def _fill_capiq_id_column(df):
    """
    NOTE: inplace
    """
    df['Blank 1'] = df['ID'].apply(id_command)

    # Blank needed because ids will populate to the right by one column
    df['IQID'] = ''


def _fill_capiq_name_column(df):
    """
    NOTE: inplace
    """
    df['Blank 2'] = df['ID'].apply(name_command)

    # Blank needed because ids will populate to the right by one column
    df['IQ Name'] = ''

def _fill_with_commands(ws, company_id: str, financial_data_items_dict: Dict[str, str],
                        market_data_items_dict: Dict[str, str], **financials_kwargs):
    """
    Note: inplace
    """

    # Set default freq
    try:
        freq = financials_kwargs['freq']
    except KeyError:
        freq = 'Q'

    date_dict = _get_date_var_dict_from_freq(freq)

    column_generator = excel_cols()

    common_args = (
        column_generator,
        ws,
        company_id
    )

    _fill_ws_by_data_item_dict(date_dict, financial_data_command, *common_args, **financials_kwargs)
    _fill_ws_by_data_item_dict(financial_data_items_dict, financial_data_command, *common_args, **financials_kwargs)
    _fill_ws_by_data_item_dict(market_data_items_dict, market_data_command, *common_args, **financials_kwargs)



def _fill_ws_by_data_item_dict(data_items_dict: Dict[str, str], command_func: Callable,
                               column_generator: Iterable, ws,
                               company_id: str, **financials_kwargs):
    for item in data_items_dict:
        current_column = next(column_generator)
        ws[f'{current_column}1'] = command_func(
            company_id,
            item,
            **financials_kwargs,
            data_item_label=data_items_dict[item]
        )


def _fill_with_holdings_commands(ws, company_id, date_str, data_items_dict):
    column_generator = excel_cols()

    for item in data_items_dict:
        current_column = next(column_generator)
        ws[f'{current_column}1'] = holdings_command(
            company_id, item, date_str,
            data_item_label=data_items_dict[item]
        )

def _get_date_var_dict_from_freq(freq) -> Dict[str, str]:
    date_dict = {
        'IQ_PERIODDATE_BS': 'Date'
    }
    if freq == 'Y':
        date_dict.update(dict(
            IQ_FISCAL_Y='Fiscal Year'
        ))
    elif freq == 'Q':
        date_dict.update(dict(
            IQ_ABS_PERIOD='Fiscal Quarter'
        ))
    else:
        raise ValueError('Must pass Y or Q for freq')

    return date_dict

def _date_str_to_file_format(date_str):
    return date_str.replace('/','-')





import datetime
import pandas as pd

CAPIQ_DATE_STR_FORMAT = '%m/%d/%Y'


def freq_and_periods_to_begin_date_str(freq: str, periods: int) -> str:
    ts = _freq_and_periods_to_begin_date(freq, periods)
    return ts.strftime(CAPIQ_DATE_STR_FORMAT)


def _freq_and_periods_to_begin_date(freq: str, periods: int) -> pd.Timestamp:
    all_dates = pd.date_range(end=datetime.datetime.today(), periods=periods, freq=freq)
    return all_dates[0]


def today_as_str() -> str:
    return datetime.datetime.today().strftime(CAPIQ_DATE_STR_FORMAT)


import os
import shutil
import time


def get_path_of_failed_folder_add_if_necessary(orig_folder):
    return _join_path_and_create_if_necessary(orig_folder, r'..\failed')

def get_path_of_additional_failed_folder_add_if_necessary(orig_folder):
    return _join_path_and_create_if_necessary(orig_folder, r'failed')

def _join_path_and_create_if_necessary(orig_folder, new_relative_path):
    out_path = os.path.join(orig_folder, new_relative_path)
    if not os.path.exists(out_path):
        os.makedirs(out_path)

    return out_path


def move_file_to_failed_folder(file, failed_folder, retries_remaining=100):
    if retries_remaining <= 0:
        print(f'ERROR SKIPPING {file}! Could not move to {failed_folder}')
        return

    try:
        shutil.copy(file, failed_folder)
    except (PermissionError, OSError):
        time.sleep(.1)
        move_file_to_failed_folder(file, failed_folder, retries_remaining=retries_remaining - 1)


import conf
from version import __version__
from setuptools import setup, find_packages


setup(
    name=conf.PACKAGE_NAME,
    version=__version__,
    description=conf.PACKAGE_SHORT_DESCRIPTION,
    long_description=conf.PACKAGE_DESCRIPTION,
    author=conf.PACKAGE_AUTHOR,
    author_email=conf.PACKAGE_AUTHOR_EMAIL,
    license=conf.PACKAGE_LICENSE,
    packages=find_packages(),
    include_package_data=True,
    classifiers=conf.PACKAGE_CLASSIFIERS,
    install_requires=conf.PACKAGE_INSTALL_REQUIRES,
    project_urls=conf.PACKAGE_URLS,
    url=conf.PACKAGE_URLS['Code']
)


from typing import List, Union
import pandas as pd
from exceldriver.columns import get_n_cols_after_col


def write_df_to_ws_values(df: pd.DataFrame, ws, begin_col: str = 'A', begin_row: int = 1):
    # TODO: multi-index
    num_cols = len(df.columns) + 1  # add 1 as index will automatically be converted to col
    end_col = get_n_cols_after_col(begin_col, num_cols - 1)  # -1 as first col goes in begin_col
    end_row = begin_row + len(df)
    cell_range = f'{begin_col}{begin_row}:{end_col}{end_row}'
    values = _df_to_values_for_insert_into_ws(df)
    ws.Range(cell_range).Value = values


def _df_to_values_for_insert_into_ws(df: pd.DataFrame) -> List[List[Union[str, float, int]]]:
    temp_df = df.reset_index()
    temp_df.fillna('', inplace=True)
    temp_df['Date'] = temp_df['Date'].apply(lambda x: x.strftime('%m/%d/%Y') if not isinstance(x, str) else x)
    values_list = temp_df.values.tolist()
    values_list.insert(0, list(temp_df.columns))
    return values_list
